// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

library editex.demo;

import 'package:eqlib/eqlib.dart';
import 'package:eqlib/latex.dart';
import 'package:editex/editex.dart';

final _defaultTemplates = {
  'a': [r'\vec{a}_c', r'\vec{a}', r'{}_\text{?}a'],
  'abs': r'\left|${0}\right|',
  'sqrt': r'\sqrt{${0}\,}',
  'root': r'\sqrt[${0}]{${1}\,}',
  'log': r'\log_{${0}}${:1(+).}',
  'sin': r'\sin${:0(+).}',
  'cos': r'\cos${:0(+).}',
  'celcius': r'${.0(*)}^\circ',
  'lim': r'\lim_{${0}\to${1}}${2(+).}',
  'diff': r'\frac{\partial}{\partial${:0(+)}}${:1(+).}',
  'frac': r'\frac{\,${0}\,}{\,${1}\,}',
  'int': r'\displaystyle\int_{${0}}^{${1}}${2}~\text{d}${:3(+).}',
  'sum': r'\displaystyle\sum_{${0}=${1}}^{${2}}${:3(+).}'
};

final Map<String, dynamic> _defaultTemplateLabels = {
  'a': ['Centripetal acceleration', 'Acceleration', 'Generic variable a'],
  'abs': 'Absolute value',
  'sqrt': 'Square root',
  'root': 'Root',
  'log': 'Logarithm',
  'sin': 'Sine',
  'cos': 'Cosine',
  'celcius': 'Degrees celcius',
  'lim': 'Limit',
  'diff': 'Partial derrivative',
  'frac': 'Fraction',
  'int': 'Definite integral',
  'sum': 'Sum series'
};

/// Test implementation of [EdiTeXInterface].
class EdiTeXTestInterface implements EdiTeXInterface {
  final ctx = new SimpleExprContext();
  final templates = new List<EdiTeXTemplate>();
  final byKeyword = new Map<String, List<EdiTeXTemplate>>();
  final byInstantChar = new Map<String, EdiTeXTemplate>();

  EdiTeXTestInterface() {
    // Initialize templates.
    _defaultTemplates.forEach((key, value) {
      if (value is String) {
        addTemplate(key, _defaultTemplateLabels[key], value);
      } else if (value is List) {
        for (var i = 0; i < value.length; i++) {
          addTemplate(key, _defaultTemplateLabels[key][i], value[i]);
        }
      }
    });

    // Initialize instant templates.
    final createTemplate = (String str, String parsableTemplate,
        [EdiTeXTT type = EdiTeXTT.other]) {
      final template = new EdiTeXTemplate(templates.length, '',
          parseLaTeXTemplate(str, ctx.operators), parsableTemplate, type);
      templates.add(template);
      return template;
    };

    byInstantChar['('] = createTemplate(r'\left(${0}\right)', r'($0)');
    byInstantChar['['] = createTemplate(r'\left[${0}\right]', r'($0)');
    byInstantChar['^'] = createTemplate(r'^{${0}}', r'^($0)', EdiTeXTT.sup);
    byInstantChar['_'] = createTemplate(r'_{${0}}', r'_($0)', EdiTeXTT.sub);
    byInstantChar['*'] = createTemplate(r'{}\cdot{}', '*');
    byInstantChar['='] = createTemplate(r'{}={}', '=', EdiTeXTT.eq);
    byInstantChar['+'] = createTemplate(r'{}+{}', '+', EdiTeXTT.add);
    byInstantChar['-'] = createTemplate(r'{}-{}', '-', EdiTeXTT.subtract);
    byInstantChar['~'] = createTemplate(r'-', '-');
    byInstantChar['!'] = createTemplate(r'!', '!');
  }

  void addTemplate(String keyword, String label, String templateSource) {
    final t = parseLaTeXTemplate(templateSource, ctx.operators);
    final args = new List<String>.generate(t.parameterCount, (i) => '\$$i');
    final type = keyword == 'frac' ? EdiTeXTT.frac : EdiTeXTT.other;
    final template = new EdiTeXTemplate(templates.length, label, t,
        args.isNotEmpty ? '$keyword(${args.join(',')})' : keyword, type);

    templates.add(template);
    byKeyword.putIfAbsent(keyword, () => new List<EdiTeXTemplate>());
    byKeyword[keyword].add(template);
  }

  @override
  EdiTeXTemplate getTemplate(id) =>
      id >= 0 && id < templates.length ? templates[id] : null;

  @override
  List<EdiTeXTemplate> lookupKeyword(keyword) {
    return byKeyword.containsKey(keyword) ? byKeyword[keyword] : [];
  }

  @override
  EdiTeXTemplate lookupCharacter(char) {
    return byInstantChar[char];
  }
}
