// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of editex;

/// Keyword used when the user types '/'. This cannot be configured. It is
/// considered a good practice to stick to LaTeX like commands when those are
/// present.
const _divideTemplateKeyword = 'frac';

/// For [EdiTeX.undoHistory] and [EdiTeX.redoHistory].
class EditHistoryEntry {
  final List data;
  final int cursor;
  EditHistoryEntry(this.data, this.cursor);
}

/// EdiTeX editor component.
/// Can initialize itself in an empty <div>
class EdiTeX {
  /// Enable or disable internal tab navigation (the user can jump to the next
  /// placeholder with tab, and to the previous one with shift-tab).
  static bool enableTabNavigation = true;
  static String selectionColor = '#2aa198';

  final EdiTeXInterface interface;
  final DivElement container, katexWrapper;

  final data = new List<dynamic>();
  var readonly = false;

  final undoHistory = new List<EditHistoryEntry>();
  final redoHistory = new List<EditHistoryEntry>();

  final _allTemplates = new Map<int, EdiTeXTemplate>();
  var _cursorIndex = 0, _selectionCursor = -1;
  var _whichTemplate = 0;
  var _doUpdate = true;
  var _wrapperWidth;
  Element _autocomplete;

  final _onLeftLeave = new StreamController<Null>.broadcast();
  final _onRightLeave = new StreamController<Null>.broadcast();
  final _onUpdate = new StreamController<Null>.broadcast();
  final _onIdleKeyDown = new StreamController<KeyboardEvent>();

  EdiTeX(this.container, this.interface)
      : katexWrapper = ht.div('.editex-wrapper') {
    container.append(katexWrapper);

    // Make focussable via keyboard.
    container.tabIndex = 0;

    // Add keyboard listeners.
    container.onKeyDown.listen(onKeyDown);

    // Clear selection cursor on blur.
    container.onBlur.listen((_) {
      if (_selectionCursor != -1) {
        _selectionCursor = -1;
        _doUpdate = true;
      }
    });

    // Update wrapperWidth on.
    _recomputeWrapperWidth();
    window.onResize.listen((_) {
      _recomputeWrapperWidth();
    });

    // Setup rendering loop.
    // Minimal interval is 33ms, this gives a framerate of ~30fps.
    new Timer.periodic(new Duration(milliseconds: 33), (t) {
      if (_doUpdate) {
        _update();
      }
    });
  }

  /// Recompute [_wrapperWidth]
  void _recomputeWrapperWidth() {
    final padding = katexWrapper.getComputedStyle().paddingLeft;
    final paddingv = num.parse(padding.substring(0, padding.length - 2));
    _wrapperWidth = katexWrapper.getBoundingClientRect().width - 2 * paddingv;
  }

  /// Destroy editor (closes broadcast streams).
  Future destroy() {
    return Future.wait([_onLeftLeave.close(), _onRightLeave.close()]);
  }

  /// Focus editor.
  void focus() => container.focus();

  /// Render equation.
  void _update() {
    // Set [doUpdate] to false and notify [_onUpdate].
    _doUpdate = false;
    _onUpdate.add(null);

    // Reset autocomplete index.
    _whichTemplate = 0;

    // Remove existing autocomplete.
    if (_autocomplete != null) {
      _autocomplete.remove();
      _autocomplete = null;
    }

    // If the user has already typed a keyword here, show an autocomplete list.
    var autocompleteAt = -1;
    final keyword = getKeyword(data, _cursorIndex);
    final autocompleteTemplates = interface.lookupKeyword(keyword);
    if (autocompleteTemplates.isNotEmpty) {
      autocompleteAt = _cursorIndex - keyword.length;
    }

    // Render content using KaTeX.
    // Note that this code throws an exception if any of the LaTeX is invalid.
    // This can for example be the case if there is a double superscript.
    katexWrapper.children.clear();
    bool rendered = false;
    while (!rendered) {
      try {
        final latex = getLaTeX(true, autocompleteAt);
        final options = new katex.RenderingOptions(displayMode: true);
        katex.render(latex, katexWrapper, options);
        rendered = true;
      } catch (e, stackTrace) {
        print('LaTeX update error: $e');
        print(stackTrace);
        if (undoHistory.isEmpty) {
          return;
        } else {
          undo(true);
        }
      }
    }

    // Format cursor and placeholders.
    Element cursor;
    num autocompleteLeft = 0;
    final specialNodes = katexWrapper.getElementsByClassName('mord mathit');
    for (var i = 0; i < specialNodes.length; i++) {
      final Element element = specialNodes[i];
      if (element.style.color == 'transparent') {
        final id = element.childNodes.first.text;
        element.childNodes.first.remove();
        if (id == 'c') {
          element.append(ht.span('.editex-cursor'));
          cursor = element;
        } else if (id == 'e') {
          element.append(ht.span('.editex-placeholder'));
        } else if (id == 'a') {
          // Compute position for autocomplete.
          final aSpan = ht.span('', attrs: {'style': 'display: inline-block;'});
          element.append(aSpan);
          final bbox = element.children.first.getBoundingClientRect();

          // Remove special node.
          element.remove();
          i--;

          // Create autocomplete element.
          var lbli = 0;
          _autocomplete = ht.div('.editex-autocomplete', c: [
            ht.div('.editex-autocomplete-popup',
                c: autocompleteTemplates.map((t) {
                  final addcss = lbli++ == _whichTemplate ? '.highlight' : '';
                  return ht.div(['.editex-autocomplete-item$addcss', t.label]);
                }).toList())
          ]);

          // Add autocomplete to container.
          final cbbox = container.getBoundingClientRect();
          autocompleteLeft = bbox.left - cbbox.left;
          _autocomplete.style.top = '${bbox.top - cbbox.top}px';
          _autocomplete.style.left = '${autocompleteLeft}px';
          container.append(_autocomplete);
        }
      }
    }

    // Compute katexDiv offset using cursor position
    final katexDisplay = katexWrapper.children.first;
    final katexLeft = katexDisplay.getBoundingClientRect().left;
    var cursorOffset = cursor.getBoundingClientRect().left;
    cursorOffset -= katexLeft - 1; // -1 because of cursor margin.

    if (cursorOffset > _wrapperWidth) {
      final offset = cursorOffset - _wrapperWidth;
      katexDisplay.style.marginLeft = '-${offset}px';

      // Move autocomplete along.
      if (_autocomplete != null) {
        autocompleteLeft = autocompleteLeft - offset;
        _autocomplete.style.left = '${autocompleteLeft}px';
      }
    } else {
      katexDisplay.style.marginLeft = '0px';
    }

    // Align autocomplete popup menu.
    if (_autocomplete != null) {
      final side =
          autocompleteLeft < container.clientWidth / 2 ? 'right' : 'left';
      _autocomplete.children.first.classes.add('editex-autocomplete-$side');
    }
  }

  Stream<Null> get onLeftLeave => _onLeftLeave.stream;
  Stream<Null> get onRightLeave => _onRightLeave.stream;
  Stream<Null> get onUpdate => _onUpdate.stream;
  Stream<KeyboardEvent> get onIdleKeyDown => _onIdleKeyDown.stream;

  bool get isEmpty => data.isEmpty;
  bool get isNotEmpty => data.isNotEmpty;
  bool get hasSelection => _selectionCursor != -1;

  /// Set cursor position to [index]. If [index] is negative it will start from
  /// the end.
  void setCursor(int index) {
    _cursorIndex = index < 0 ? data.length + index + 1 : index;
    _selectionCursor = -1;
    _doUpdate = true;
  }

  /// Add undo checkpoint.
  void _addHistoryEntry([EditHistoryEntry entry]) {
    undoHistory.add(entry ?? getState());
    redoHistory.clear();
  }

  /// Get current editor state.
  EditHistoryEntry getState() =>
      new EditHistoryEntry(new List.from(data), _cursorIndex);

  /// Undo last change.
  void undo([bool destructive = false]) {
    if (undoHistory.isNotEmpty) {
      if (!destructive) {
        redoHistory.add(getState());
      }

      final replace = undoHistory.removeLast();
      data
        ..clear()
        ..addAll(replace.data);
      _cursorIndex = replace.cursor;
      _doUpdate = true;
    }
  }

  /// Redo last change.
  void redo() {
    if (redoHistory.isNotEmpty) {
      undoHistory.add(getState());
      final replace = redoHistory.removeLast();
      data
        ..clear()
        ..addAll(replace.data);
      _cursorIndex = replace.cursor;
      _doUpdate = true;
    }
  }

  /// Remove selection.
  /// Returns removed items if any.
  List _removeSelection() {
    if (hasSelection) {
      final a = _cursorIndex, b = _selectionCursor;
      final removed = data.sublist(a < b ? a : b, a < b ? b : a);
      data.removeRange(a < b ? a : b, a < b ? b : a);
      _cursorIndex = a < b ? a : b;
      _selectionCursor = -1;
      return removed;
    } else {
      return [];
    }
  }

  /// Get selection as JSON data.
  String get selectionJson {
    final a = _cursorIndex, b = _selectionCursor;
    return JSON.encode(data.sublist(a < b ? a : b, a < b ? b : a));
  }

  /// Keyboard handler.
  void onKeyDown(KeyboardEvent event) {
    // Save current editor state.
    final oldState = getState();

    // Set selection cursor to -1 if it is at the same index as the main cursor.
    if (_selectionCursor == _cursorIndex) {
      _selectionCursor = -1;
    }

    // Keystrokes
    if (event.ctrlKey) {
      if (event.keyCode == KeyCode.Z) {
        if (event.shiftKey) {
          redo();
        } else {
          undo();
        }

        // To prevend the final code to mess up undo/redo, handle this here.
        _doUpdate = true;
        event.preventDefault();
        return;
      } else if (event.keyCode == KeyCode.C) {
        if (hasSelection) {
          window.localStorage['editex-clipboard'] = selectionJson;
        }
      } else if (event.keyCode == KeyCode.V) {
        _removeSelection();
        final clipboardDataStr = window.localStorage['editex-clipboard'];
        if (clipboardDataStr != null) {
          final List clipboardData = JSON.decode(clipboardDataStr);
          _loadTemplates(clipboardData);
          data.insertAll(_cursorIndex, clipboardData);
          _cursorIndex += clipboardData.length;
        }
      }
    }

    // Tab navigation
    else if (enableTabNavigation && event.keyCode == KeyCode.TAB) {
      _completeKeyword();

      final back = event.shiftKey;
      final placeholders = getEmptyArguments(data, _allTemplates);
      final c = _cursorIndex;
      _cursorIndex = placeholders.reduce((a, b) {
        // Notice that we know that a < b by order of iteration.
        return back
            ? (b < c ? b : (a < c ? a : c))
            : (a > c ? a : (b > c ? b : c));
      });
    }

    // Selection
    else if (event.shiftKey &&
        (event.keyCode == KeyCode.LEFT || event.keyCode == KeyCode.RIGHT)) {
      if (_selectionCursor == -1) {
        _selectionCursor = _cursorIndex;
      }

      _selectionCursor = moveSelectionCursor(data, _selectionCursor,
          event.keyCode == KeyCode.LEFT ? -1 : 1, _allTemplates);

      // Selection cursor changes are not recognized as state change.
      _doUpdate = true;
      event.preventDefault();
      return;
    }

    // Normal insert
    else if (new RegExp(r'^[0-9A-Za-z]$').hasMatch(event.key)) {
      var char = event.key;
      if (event.shiftKey) {
        char = char.toUpperCase();
      }

      _removeSelection();
      data.insert(_cursorIndex++, char);
    }

    // Insert backward slash (LaTeX command separator).
    else if (event.key == r'\') {
      _removeSelection();
      if (_completeKeyword()) {
        data.insert(_cursorIndex++, r'\');
      }
    }

    // Instant template insert
    else if (interface.lookupCharacter(event.key) != null) {
      _completeKeyword();
      final template = interface.lookupCharacter(event.key);

      // If there is a selection, save it and add it to the first argument.
      if (hasSelection) {
        final selection = _removeSelection();
        insertTemplateAtCursor(template);
        if (template.parameterCount > 0) {
          data.insertAll(_cursorIndex, selection);
          _cursorIndex += selection.length + 1;
        }
      } else {
        insertTemplateAtCursor(template);
      }
    }

    // Special divide behavior
    else if (event.key == '/') {
      // Only proceed if [_divideTemplateKeyword] is specified.
      final fracTemplates = interface.lookupKeyword(_divideTemplateKeyword);
      if (fracTemplates.isNotEmpty) {
        _completeKeyword();

        // If no selection is present, select everything before this until the
        // first =+- template.
        if (_selectionCursor == -1) {
          // Get breaking template IDs.
          final breakAt = new Set<int>.from(['=', '+', '-']
              .map((char) => interface.lookupCharacter(char).id)
              .toList());

          // Start at current cursor index and iterate backwards.
          _selectionCursor = _cursorIndex;

          // Stop iteration at index 0 or when the next (-1) element is a
          // breaking template ID.
          while (_selectionCursor > 0 &&
              (_selectionCursor == data.length ||
                  !breakAt.contains(data[_selectionCursor - 1]))) {
            // Move one space back.
            final newIndex =
                moveSelectionCursor(data, _selectionCursor, -1, _allTemplates);

            // If the newIndex is the same as the old one we hit another
            // boundary, so it's time to break the loop.
            if (newIndex != _selectionCursor) {
              _selectionCursor = newIndex;
            } else {
              break;
            }
          }
        }

        // Remove selection.
        final numerator = _removeSelection();

        // Insert fraction.
        insertTemplateAtCursor(fracTemplates.first);

        // Copy selection to first argument and proceed to second argument.
        data.insertAll(_cursorIndex, numerator);
        _cursorIndex += numerator.length + 1;
      }
    }

    // Finish a keyword.
    else if (event.keyCode == KeyCode.SPACE || event.keyCode == KeyCode.ENTER) {
      _completeKeyword();
    }

    // Remove before cursor.
    else if (event.keyCode == KeyCode.BACKSPACE) {
      if (_removeSelection().isEmpty && _cursorIndex > 0) {
        _cursorIndex = removeElement(data, _cursorIndex - 1, _allTemplates, -1);
      }
    }

    // Remove after cursor.
    else if (event.keyCode == KeyCode.DELETE) {
      if (_removeSelection().isEmpty && _cursorIndex < data.length) {
        _cursorIndex = removeElement(data, _cursorIndex, _allTemplates, 1);
      }
    }

    // Arrow key navigation.
    else if (event.keyCode == KeyCode.LEFT) {
      //_completeKeyword();
      if (_selectionCursor != -1) {
        setCursor(min(_selectionCursor, _cursorIndex));
      } else if (--_cursorIndex < 0) {
        _cursorIndex = 0;
        _onLeftLeave.add(null);
      }
    } else if (event.keyCode == KeyCode.RIGHT) {
      //_completeKeyword();
      if (_selectionCursor != -1) {
        setCursor(max(_selectionCursor, _cursorIndex));
      } else if (++_cursorIndex > data.length) {
        _cursorIndex = data.length;
        _onRightLeave.add(null);
      }
    }

    // Home/End navigation
    else if (event.keyCode == KeyCode.HOME) {
      //_completeKeyword();
      setCursor(0);
    } else if (event.keyCode == KeyCode.END) {
      //_completeKeyword();
      setCursor(data.length);
    }

    // Autocomplete navigation
    else if (event.keyCode == KeyCode.UP || event.keyCode == KeyCode.DOWN) {
      // No new render cycle will be triggered, modify the autocomplete here.
      if (_autocomplete != null) {
        event.preventDefault();

        final items = _autocomplete.children.first.children;
        items[_whichTemplate].classes.remove('highlight');

        _whichTemplate += event.keyCode == KeyCode.DOWN ? 1 : -1;
        _whichTemplate = _whichTemplate < 0 ? items.length - 1 : _whichTemplate;
        _whichTemplate = _whichTemplate == items.length ? 0 : _whichTemplate;

        items[_whichTemplate].classes.add('highlight');
      } else {
        return;
      }
    }

    // Check if the editor state is updated.
    final contentUpdated = !(const ListEquality().equals(oldState.data, data));
    final cursorMoved = oldState.cursor != _cursorIndex;

    if (contentUpdated) {
      if (readonly) {
        // This it a bit hacky, but very lightweight.
        setCursor(oldState.cursor);
        data.clear();
        data.addAll(oldState.data);
      } else {
        _addHistoryEntry(oldState);
        _doUpdate = true;
        event.preventDefault();
      }
    } else if (cursorMoved) {
      _doUpdate = true;
      event.preventDefault();
    } else {
      _onIdleKeyDown.add(event);
    }
  }

  /// Complete keyword.
  /// Returns true if a valid keyword was found and resolved or if there is no
  /// keyword at the current position (or '\').
  bool _completeKeyword([bool addHistoryEntry = false]) {
    final keyword = getKeyword(data, _cursorIndex);
    if (keyword.isEmpty) {
      return true;
    }

    // Lookup template.
    final templates = interface.lookupKeyword(keyword.replaceAll(r'\', ''));
    if (templates.isNotEmpty) {
      // Remove fragments that form the keyword.
      data.removeRange(_cursorIndex - keyword.length, _cursorIndex);
      _cursorIndex -= keyword.length;

      // Insert the template.
      assert(_whichTemplate < templates.length);
      insertTemplateAtCursor(templates[_whichTemplate]);

      // Add history entry.
      if (addHistoryEntry) {
        _addHistoryEntry();
      }

      return true;
    } else {
      return false;
    }
  }

  /// Insert given template at cursor.
  void insertTemplateAtCursor(EdiTeXTemplate template) {
    _allTemplates[template.id] = template;
    _cursorIndex = insertTemplate(data, _cursorIndex, template);
  }

  /// Get JSON serializable data object.
  List getData() => data;

  /// Load data returned by [getData].
  void loadData(List src) {
    _addHistoryEntry();
    _loadTemplates(src);

    data.clear();
    data.addAll(src);

    _cursorIndex = 0;
    _doUpdate = true;
  }

  /// Make sure all templates in [src] are loaded.
  void _loadTemplates(List src) {
    for (final value in src) {
      if (value is int && value >= 0 && !_allTemplates.containsKey(value)) {
        _allTemplates[value] = interface.getTemplate(value);
      }
    }
  }

  /// Get content as parsable expression.
  String getParsable() {
    return generateString(false, data, _allTemplates);
  }

  /// Get LaTeX string.
  String getLaTeX([bool includeCursors = false, int autocompleteAt = -1]) {
    final cursors = new List<int>();
    if (includeCursors) {
      cursors.add(_cursorIndex);
    }
    if (_selectionCursor != -1 && _selectionCursor != _cursorIndex) {
      cursors.add(_selectionCursor);
    }

    return generateString(
        true, data, _allTemplates, cursors, selectionColor, autocompleteAt);
  }
}
